use strict;
package europa_module_factory;

use vars qw ($AUTOLOAD);

my $db = {};

# Object constructor.
sub new
{
  my $class = shift;
  my $options = shift;

  my $this = $options;
  bless($this, $class);

  my $first_module_of_its_class = 0;
  my $db_key = ref $this;
  if (!exists $db->{$db_key})
  {
    $db->{$db_key} = {};
    $db->{$db_key}->{fields} = $class->get_fields();
    $first_module_of_its_class = 1;
  }

  # Call setters with each option.
  for my $option (keys %$options)
  {
    my $value = $options->{$option};
    $this->$option($value);
  }

  # Verify that required fields were supplied.
  my $fields = $class->get_fields();
  for my $field (keys %{$fields})
  {
    if (!$fields->{$field}->{optional})
    {
      if (!defined $this->$field())
      {
        print STDERR "$class: required option '$field' is not defined\n";
        exit 1;
      }
    }
  }

  if (!defined $this->module())
  {
    # If no name is assigned, come up with a unique one based
    # on $class and options.  But if this is the first module
    # of its class (which may be a common case) take a shortcut -
    # choose a simple name.
    if (!defined $this->{name})
    {
      my $name = $class;
      $name =~ s/::/_/g;
      
      if (!$first_module_of_its_class)
      {
        $name .= "_" if (keys %$this);
        $name .= join("_", (map {"$_\_$this->{$_}"} keys %$this));
      }
      $this->{name} = $name;
    }
    my $module = e_module->new({name => $this->{name}});

    $this->module($module);
    $this->document();

    $this->add_contents_to_module();
  }

  return $this;
}

# AUTOLOAD setters and getters.
# If a fields hash is defined for the option name,
# it's used to verify the value.  No fields hash means
# anything goes... may want to make that be a failure instead.
sub AUTOLOAD
{
  my $this = shift;
  my $value = shift;
  my $name = $AUTOLOAD;

  # Strip off all but the last element of $name.  Separator is "::".
  my @parts = split('::', $name);
  $name = $parts[$#parts];

  # This method can be called statically or with an actual object.
  my $class = $this;
  $class = ref($class) if ref($class);

  if (defined $value)
  {
    # Look up this class's option validations.
    my $fields = $db->{$class}->{fields};
    if (!$class->validate($name, $value, $fields->{$name}))
    {
      exit 1;
    }
    if (ref $this)
    {
      $this->{$name} = $value;
    }
  }
  return $this->{$name};
}

# Subclass should override this method to add module contents.
# If the method is not overridden, an empty stub module is created.
#
# To do:Instances of the empty module appear to be empty - it would be 
# nice if they showed up, even if their existence meant bogus HDL.
sub add_contents_to_module
{
  my $this = shift;
  my $module = $this->module();

  my $existing_comment = $module->comment();
  $module->comment(
    $existing_comment . 
    "\nThis is a stub module.\nAdd contents to this module by overriding\n" . 
    ref($this) .
    "::add_contents_to_module().\n"
  );
}

sub validate
{
  my $class = shift;
  my $name = shift;
  my $value = shift;
  my $validator = shift;
  my $retval = 1;

  return 1 if (!defined $validator);
  my ($min, $max) = @{$validator->{range}};

  if (defined($min) && $value < $min)
  {
    print STDERR "$class\::validate(): " .
      "value '$value' for option '$name' is less than " .
      "allowed min value '$min'\n";
    $retval = 0;
  }
  if (defined($max) && $value > $max)
  {
    print STDERR "$class\::validate(): " .
      "value '$value' for option '$name' is greater than " .
      "allowed max value '$max'\n";
    $retval = 0;
  }

  return $retval;
}

sub get_existing_module
{
  my $this = shift;
  my $module = undef;

  my $key = $this->compute_module_key();
  if (exists $db->{ref $this}->{$key})
  {
    $module = $db->{ref $this}->{$key};
  }

  return $module;
}

sub compute_module_key
{
  my $this = shift;

  my $key = join("", (map {"$_=>$this->{$_},"} sort keys %$this));
  return $key;
}

sub module
{
  my $this = shift;
  if (@_) 
  {
    my $key = $this->compute_module_key();
    $db->{ref $this}->{$key} = shift;
  }

  return $this->get_existing_module();
}

sub document
{
  my $this = shift;

  my $comment;
  # Put the name of the instancing class right at the top, for reference.
  $comment .= ("Module class: " . ref($this) . "\n");
  $comment .= "Module options:\n";

  for my $key (sort keys %$this)
  {
    my $value = $this->{$key};
    $value = "<ref>" if ref $value;
    $comment .= "$key: $value\n";
  }

  my $module = $this->module();
  if (defined $module)
  {
    $this->module()->comment($comment);
  }
}

sub make_instance
{
  my $this = shift;

  my $options = shift;

  my $module = $this->get_existing_module();
  $options->{module} = $module;

  my $instance = undef;
  if (defined $module)
  {
    $instance = e_instance->new($options);
  }
  else
  {
    print STDERR __PACKAGE__ . "::make_instance(): Error: non-existent module\n";
  }

  return $instance;
}

1;

